/** @file
 ** Main program, unique file in the project.
 *
 * This is a replacement for notepad.exe or other programs that are
 * inconditionaly launched by another program.
 * It's aim is to replace the binary of the program that is ran,
 * and it will launch the program of the user's choice,
 * defined in the <ProgramName>.cfg file.
 * In this file, the first line is the path to the program to launch.
 * The second one gives the parameters to pass to the program, including,
 * the command line given to this program (ie. probably the path of the
 * file to open). Just put an '*' where this command line
 * string must be inserted.
 * The third line is the path of the launch directory (Start in).
 * If this line begins with a dot, this dot will be replaced by the
 * current working directory.
 **/
/* Project history:
 *  1.01.001  2002/07/22 (PL)  Better error handling (errorlevel).
 *  1.01.000  2002/01/16 (PL)  Manage dot in third line, option switches in command line.
 *             Plus fixes of stupid bugs (like no initialization of some variables...).
 *  1.00.000  2000/12/21 (PL)  Creation
 */
/* Copyright notice: See the PhiLhoSoft Licence.txt file for details.
ProgLauncher is distributed under the zlib/libpng license.
Copyright (c) 2000-2002 Philippe Lhoste / PhiLhoSoft
*/

/*------------------*
 * Files to include *
 *------------------*/

#define STRICT	// Some rigor doesn't hurt!
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <stdio.h>

#define INIT	// Initialize global variables
#define EXT 	// Global variables are defined, not just declared
#include "resource.h"

/*-------------------*
 * Macro definitions *
 *-------------------*/

/*------------------*
 * Global variables *
 *------------------*/

const char *g_szAppName = "ProgLauncher";	// Application name
// These variables are defined globally in WUtil.h, suppress these if this file is included
HWND g_hMainWin;
HINSTANCE g_hInst;

/*--------------------*
 * Internal functions *
 *--------------------*/

/**********************************************************************
 * Windows main procedure.
 * @return Code for Windows
 **********************************************************************/
int WINAPI WinMain(
	HINSTANCE hInstance,	///< Instance handle
	HINSTANCE hPrevInst,	///< Sibbling application handle
	LPSTR  lpCmdLine,		///< Command line used to run this application
	int    nCmdShow)		///< Default display mode of the window
{
	int errorLevel = 0;
	int resSE, len;
	FILE *f;
	char sProgPath[MAX_PATH + 1], *pParams, sParamsFormat[MAX_PATH + 1];
	char sCmd[MAX_PATH + 1], sParams[MAX_PATH * 2 + 1], sDir[MAX_PATH + 1];
	char *pRes;

	UNREFERENCED_PARAMETER(hPrevInst);
	if (lpCmdLine == NULL || *lpCmdLine == '\0')	// Bad call
	{
		errorLevel = 255;	/* Bad parameter */
		goto end;
	}
	g_hInst = hInstance;	// Save our instance handle (not used!)

	sParamsFormat[0] = sCmd[0] = sParams[0] = sDir[0] = '\0';
	len = GetModuleFileName(NULL, sProgPath, MAX_PATH);
	strcpy(sProgPath + len - 3, "cfg");	/* Replace .exe by .cfg */
	f = fopen(sProgPath, "rt");
	if (f != NULL)
	{
		/* Line 1: Read the path of the executable to launch */
		pRes = fgets(sCmd, MAX_PATH, f);
		if (pRes != NULL)
		{
			len = strlen(sCmd) - 1;
			if (len >= 0 && sCmd[len] == '\n')
			{
				sCmd[len] = '\0';	/* Remove '\n' */
			}
		}
		/* Line 2: Read the parameters of the above program */
		pRes = fgets(sParamsFormat, MAX_PATH, f);
		if (pRes != NULL)
		{
			len = strlen(sParamsFormat) - 1;
			if (len >= 0 && sParamsFormat[len] == '\n')
			{
				sParamsFormat[len] = '\0';	/* Remove '\n' */
			}
		}
		/* Line 3: Read the path of the launch directory */
		pRes = fgets(sDir, MAX_PATH, f);
		if (pRes != NULL)
		{
			len = strlen(sDir) - 1;
			if (len >= 0 && sDir[len] == '\n')
			{
				sDir[len] = '\0';	/* Remove '\n' */
			}
			if (sDir[0] == '.')
			{
				/* Starts with a dot, replace it with the current working directory */
				char sCwd[MAX_PATH + 1];
				GetCurrentDirectory(MAX_PATH + 1, sCwd);
				len = strlen(sCwd) - 1;
				if (len >= 0 && sCwd[len] == '\\')
				{
					sCwd[len] = '\0';	/* Remove trailing backslash (if at root dir) */
				}
				strcat(sCwd, sDir + 1);
				/* Copy back result */
				strcpy(sDir, sCwd);
			}
		}
		fclose(f);
	}
	else
	{
		errorLevel = 255;	/* Bad parameter */
		goto end;
	}
	len = 0;
	while (*lpCmdLine == '/' || *lpCmdLine == '-')
	{
		/* Copy the option (space/tab delimited) in the parameters */
		do
		{
			sParams[len++] = *lpCmdLine++;
		} while (*lpCmdLine != '\0' && *lpCmdLine != ' ' && *lpCmdLine != '\t');
		/* Skip (and copy) spaces */
		while (*lpCmdLine != '\0' && (*lpCmdLine == ' ' || *lpCmdLine == '\t'))
		{
			sParams[len++] = *lpCmdLine++;
		}
	}
	if (*lpCmdLine == '"')
	{
		// Strip the double-quotes given in the command line
		// because we are likely to add them from the format given
		int l;
		lpCmdLine++;
		l = strlen(lpCmdLine);
		if (l > 0 && lpCmdLine[l - 1] == '"')
		{
			lpCmdLine[l - 1] = '\0';
		}
	}

	// Parse the params line, to put the parameter given to this program in the right place
	pParams = strchr(sParamsFormat, '*');	// '*' is where the cmdLine is put
	if (pParams)
	{
		*pParams = '\0';
		strcat(sParams, sParamsFormat);
		strcat(sParams, lpCmdLine);
		strcat(sParams, pParams + 1);
	}
	else	// No '*', put it in double-quotes at the end
	{
		strcat(sParams, sParamsFormat);
		strcat(sParams, " \"");
		strcat(sParams, lpCmdLine);
		strcat(sParams, "\"");
	}
	resSE = (int)ShellExecute(
				NULL,			// Parent window for message boxes during app. start
				"open",			// The verb giving the action to perform
				sCmd,			// File to open
				sParams,		// Parameters
				sDir,			// Launch directory
				nCmdShow);		// Show command: it the one given to this program
	if (resSE <= 32)
	{
		// Trouble! We should perhaps handle errors...
		errorLevel = resSE;
		goto end;
	}

end:
	if (errorLevel > 0)
	{
		exit(errorLevel);
	}
	return 0;
}
