-- SciTEStartup.lua an initialisation script for SciTE
-- Ctrl+Shift+O on line below (ajust to your path...) to get the list of Scintilla functions.
-- C:\Dev\SciTE\_current\Scintilla\include\Scintilla.iface

--------------------------------------------------------------------------------
--- Generic useful functions ---

-- Cut a string in several parts, along the given separator (only one char)
function Split(toCut, separator)
	local splitted = {}
	local i = 0
	-- Regular expression:
	-- 0 or more chars different from separator, followed by 0 or 1 separator
	local regEx = "([^" .. separator .. "]*)" .. separator .. "?"
	-- Add a separator at the end, to take in account empty field at the end of the line
	for item in string.gfind(toCut .. separator, regEx) do
		i = i + 1
		splitted[i] = item or ''
	end
	splitted[i] = nil
	return splitted
end

function RepeatEditor(f, n)
	for i = 1, n do
		editor[f](editor)
	end
end

-- Get EOL char(s), assuming consistency in the file
function GetEOL()
	local eol = "\r\n"
	if editor.EOLMode == SC_EOL_CR then
		eol = "\r"
	elseif editor.EOLMode == SC_EOL_LF then
		eol = "\n"
	end
	return eol
end
function GetEOL()
	local map =
	{
		SC_EOL_CRLF = "\r\n",
		SC_EOL_CR = "\r",
		SC_EOL_LF = "\n",
	}
	return map[editor.EOLMode]
end
function GetEOL()
	local map =
	{
		[SC_EOL_CRLF] = "SC_EOL_CRLF",
		[SC_EOL_CR] = "SC_EOL_CR",
		[SC_EOL_LF] = "SC_EOL_LF",
	}
	return map[editor.EOLMode]
end

-- From Neil Hodgson
function CenterOnLine(line)
	local topLineWanted = editor:VisibleFromDocLine(line) - editor.LinesOnScreen / 2
	editor:LineScroll(0, topLineWanted - editor.FirstVisibleLine)
end

function GetCurrentWord()
	editor:WordRight()
	editor:WordLeftExtend()
	return editor:GetSelText()
end

function GetLineData(lineNb)
	if lineNb < 0 then
		lineNb = editor:LineFromPosition(editor.CurrentPos)
	end
	local lineStartPos = editor:PositionFromLine(lineNb)
	local lineEndPos = editor.LineEndPosition[lineNb]
	return lineStartPos, lineEndPos, lineNb
end

function GetCurrentLine()
	local lineStartPos, lineEndPos = GetLineData(-1)
	return editor:textrange(lineStartPos, lineEndPos)
end

function SelectCurrentLine()
	local lineStartPos, lineEndPos = GetLineData(-1)
	editor:SetSel(lineStartPos, lineEndPos)
end

function ReplaceLine(lineNb, line, bUndo)
	local lineStartPos, lineEndPos = GetLineData(lineNb)
	if bUndo then
		editor:BeginUndoAction()
	end
	editor:SetSel(lineStartPos, lineEndPos)
	-- Replace line with result
	editor:ReplaceSel(line)
	-- Operation can be cancelled with a single Undo
	if bUndo then
		editor:EndUndoAction()
	end
end

--------------------------------------------------------------------------------
--- Tool functions ---

function AddSoftBrackets()
	local text, len = editor:GetSelText()
	if len == 0 then return end
	local eol = GetEOL()
	-- Get indentation level
	-- I could use GetLineIndentation...
	-- I assume code uses tabs for indentation...
	local _, _, indent = string.find(text, "^(\t+)")
	-- I suppose selected text is already at the right identation
	indent = string.sub(indent, 2)
	text = indent .. '{' .. eol .. text ..  indent .. '}' .. eol
	editor:ReplaceSel(text)
end

--~ tagStart = [[<label for="">]];
--~ tagEnd = [[</label>]];
tagStart = [[<h3><a name="|" id="|">]];
tagEnd = [[</a></h3>]];

function AddXMLTag()
	local text, len = editor:GetSelText()
	if len == 0 then
		text, len = GetCurrentWord()
	end
	local ts = string.gsub(tagStart, "|", text);
	editor:ReplaceSel(ts .. text .. tagEnd)
end

--------------------------------------------------------------------------------
--- SVG functions ---
--[[
-- Reduce floating point precision of coordinate pairs in an SVG file
function ReduceFPP2()
	-- Read current selection
	local text, len = editor:GetSelText()
	-- Search FP number pairs separated by comas
	-- and provide them to a function that round them
	-- and spit them out to replace orignal values
	text = string.gsub(text, "([.%d]+),([.%d]+)",
		function (n1, n2)
			return math.floor(n1*10 + 0.5)/10 .. ',' ..
						 math.floor(n2*10 + 0.5)/10
		end)
	-- Replace current selection with modified text
	editor:ReplaceSel(text)
end

-- Reduce floating point precision of coordinates in an SVG file
function ReduceFPP()
	-- Read current selection
	local text, len = editor:GetSelText()
	-- Search FP numbers
	-- and provide them to a function that round them
	-- and spit them out to replace orignal values
	text = string.gsub(text, "([.%d]+)",
		function (n)
			return math.floor(n*10 + 0.5)/10
		end)
	-- Replace current selection with modified text
	editor:ReplaceSel(text)
end

-- Find min and max of coordinates in an SVG file
function FindMinMax()
	local min, max = 1000000000, 0
	-- Read current selection
	local text, len = editor:GetSelText()
	-- Search FP numbers
--	local inb = 0
	for n in string.gfind(text, "([.%d]+)") do
		local nn = tonumber(n)
		if nn ~= nil then
--		inb = inb + 1
			if nn > max then max = nn end
			if nn < min then min = nn end
--print(inb, nn, min, max)
		end
	end
	-- Display result
	print(min, max)
end
]]
--------------------------------------------------------------------------------
--- Various functions ---

function Sort2(el1, el2)
	if el1[2] == '' and el2[2] == '' then
		return false
	end
	if el1[2] == '' then
		return false
	end
	if el2[2] == '' then
		return true
	end
	-- Convert to number to avoid having 13 > 111...
	return tonumber(el1[2]) < tonumber(el2[2])
end

-- Sort a file made of tab-separated fields.
-- Sort along the FIELD_TO_SORT column.
-- The Sort function contains the real comparison.
function SortTable()
	local FIELD_TO_SORT = 1
	local lineNb = editor.LineCount
	local firstLine = editor:GetLine(0)
	local toSort = {}
	local line, parts
	-- Skip first line, if it is the title line
	-- Skip last line, it should be empty, because all lines must be ended with newline char(s)
	for i = 1, lineNb - 2 do
		line = editor:GetLine(i)
		if line ~= nil then
			parts = Split(line, '\t')
			-- Store line number, key and real line
			-- Line number isn't used here...
			-- Perhaps we can avoid to store line, and swap
			-- the lines according to old line number.
			-- It would use less memory but would be probably slower.
			toSort[i] = { i, parts[FIELD_TO_SORT], line }
		end
	end
	-- Do the sort!
	table.sort(toSort, Sort2)
	-- Replace whole text with result
	editor:BeginUndoAction()
	-- Overwrite buffer content with first line (title)
	editor:SetText(firstLine)
	-- Append each line to the end of the buffer
	for i = 1, lineNb - 2 do
--~ 		editor:InsertText(editor.Length, toSort[i][3])
		editor:AppendText(toSort[i][3])
	end
	-- Sort can be cancelled with a single Undo
	editor:EndUndoAction()
end

function Sort2(el1, el2)
	if el1[2] == '' and el2[2] == '' then
		return false
	end
	if el1[2] == '' then
		return false
	end
	if el2[2] == '' then
		return true
	end
	-- Convert to number to avoid having 13 > 111...
	return tonumber(el1[2]) < tonumber(el2[2])
end

-- Sort lines.
function SortLines()
	local lineNb = editor.LineCount
	local toSort = {}
	local line, parts
	-- Skip last line, it should be empty, because all lines must be ended with newline char(s)
	for i = 0, lineNb - 2 do
		line = editor:GetLine(i)
		if line ~= nil then
			toSort[i + 1] = line
		end
	end
	-- Do the sort!
	table.sort(toSort, function (e1, e2) return e1 < e2 end )
	-- Replace whole text with result
	editor:BeginUndoAction()
	-- Overwrite buffer content with empty string
	editor:SetText("")
	-- Append each line to the end of the buffer
	for i = 0, lineNb - 2 do
--~ 		editor:InsertText(editor.Length, toSort[i][3])
		editor:AppendText(toSort[i])
	end
	-- Sort can be cancelled with a single Undo
	editor:EndUndoAction()
end

function CapitalizeLines()
	local lineNb = editor.LineCount
	local line, substNb
	editor:BeginUndoAction()
	-- Skip last line, it should be empty, because all lines must be ended with newline char(s)
	for i = lineNb - 2, 0, -1 do
		line = editor:GetLine(i)
		if line ~= nil then
			line, substNb = string.gsub(line, "([A-Z-])([A-Z-]+[ ,;-])",
					function (c, s)
						return c .. string.lower(s)
					end)
			line = string.gsub(line, "[\r\n]", "");
			ReplaceLine(i, line, false)
		end
	end
	editor:EndUndoAction()
end

-- Zen Cart French Translation Utility
function LowerCaseWords()
	local line = GetCurrentLine()
	-- Trim initial space
--	line, substNb = string.gsub(line, "^%s+define(.*)", "define%1")
	line, substNb = string.gsub(line, "([^.][ ;][A-Z])([^A-Z])", function (c, s) return string.lower(c) .. s end)
	if substNb > 0 then
		ReplaceLine(editor:LineFromPosition(editor.CurrentPos), line, true)
	else
		output:AppendText(line)
	end
	editor:CharRight()
end

-- Trim all spaces and split line in fixed length chunks
function Normalize()
	local chunkLength = 5
	local line = GetCurrentLine()
	-- Trim all spaces
	line = string.gsub(line, "%s", '')
	-- Make a regexp matching n chars (Lua's regexp are limited...)
	matcher = '(' .. string.rep('.', chunkLength) .. ')'
	-- Add a space every n chars
	line = string.gsub(line, matcher,
			function (chunk) return chunk .. ' ' end)
	-- Previous gsub may have added an extra space at end
	line = string.gsub(line, "%s$", '')
	ReplaceLine(editor:LineFromPosition(editor.CurrentPos), line, true)
end

-- Toggle boolean values
-- DaveMDS and Steve Donovan
--[[
local replacements = {
   ["TRUE"] = "FALSE", ["FALSE"] = "TRUE",
   ["YES"]="NO", ["NO"]="YES",
   ["0"]="1", ["1"]="0",
}

function Toggle()
    local StartPos = editor.CurrentPos
    editor:WordRight()
    editor:WordLeftExtend()
    local Word = editor:GetSelText()
    local was_lowercase = string.find(Word,"^[a-z]")
    local repl = replacements[string.upper(Word)]
    if was_lowercase then
        repl = string.lower(repl)
    end
    if repl then
        editor:ReplaceSel(repl)
    end
    editor:GotoPos(StartPos)
end
]]


function MakeUpperLower(s)
	return string.upper(string.sub(s, 1, 1)) ..
           string.lower(string.sub(s, 2))
end

function PLToggle()
	local replacements =
	{
		{ "true",  "false" },
		{ "yes",   "no" },
		{ "1",     "0" },
		{ "right", "wrong" },
	}

	local startPos = editor.CurrentPos
	local word = GetCurrentWord()
	local case = "lower"
	if string.find(word,"[%u][%u]+") ~= nil then
		case = "upper"
	elseif string.find(word,"[%u][%l]+") ~= nil then
		case = "mixed"
	end
	local lword = string.lower(word)
	local repl
	for i, v in ipairs(replacements) do
		if lword == v[1] then repl = v[2]
		elseif lword == v[2] then repl = v[1]
		end
	end
	if repl then
		if case == "upper" then
			repl = string.upper(repl)
		elseif case == "mixed" then
			repl = MakeUpperLower(repl)
		end
		editor:ReplaceSel(repl)
	end
	editor:GotoPos(startPos)
end

function ChangeLineNumberStyle()
--~ 	editor.StyleFore(STYLE_LINENUMBER, 1122884) -- 0x112244
--~ 	editor.StyleBack(STYLE_LINENUMBER, 16746564) -- 0xFF8844
	editor.StyleFore[STYLE_LINENUMBER] = 1122884 -- 0x112244
	editor.StyleBack[STYLE_LINENUMBER] = 16746564 -- 0xFF8844
	print("Fait !")
end

function StyleLine()
	local H = function (h) return tonumber(h, 16) end
	local style1, style2 = 1, 2
	local STYLE_LENGTH = 10
	local lineStartPos, lineEndPos = GetLineData(-1)
	editor:StartStyling(lineStartPos, 31)
	for i = 0, (lineEndPos - lineStartPos) / STYLE_LENGTH do
		editor:SetStyling(STYLE_LENGTH, math.mod(i, 2) == 0 and style1 or style2)
	end
	editor.StyleFore[style1] = H'112288'
	editor.StyleFore[style2] = H'44AA66'
	editor.StyleBack[style1] = H'AAFF88'
	editor.StyleBack[style2] = H'AA88FF'
end

--------------------------------------------------------------------------------
--- Functions called by keyboard shortcuts (tests, etc.) ---

function Whatever(p)
--~ 	local text, len = editor:GetSelText()
--~ 	if len == 0 then return end
	local eol = GetEOL()
	print(({
		[SC_EOL_CRLF] = "SC_EOL_CRLF",
		[SC_EOL_CR] = "SC_EOL_CR",
		[SC_EOL_LF] = "SC_EOL_LF",
	})[editor.EOLMode])
	-- Loop on the lines and process them
--~ 	text = string.gsub(text, ".-" .. eol, function (l) return '\t' .. l end)
--~ 	editor:ReplaceSel(text)
end

function QuickNDirty()
--~ 	LowerCaseWords()
--~ 	CapitalizeLines()
	Normalize()
end

function QuickNDirty_()
	local lexers =
	{
		[1] = "NULL",
		[2] = "PYTHON",
		[3] = "CPP",
		[4] = "HTML",
		[5] = "XML",
		[6] = "PERL",
		[7] = "SQL",
		[8] = "VB",
		[9] = "PROPERTIES",
		[10] = "ERRORLIST",
		[11] = "MAKEFILE",
		[12] = "BATCH",
		[13] = "XCODE",
		[14] = "LATEX",
		[15] = "LUA",
		[16] = "DIFF",
		[17] = "CONF",
		[18] = "PASCAL",
		[19] = "AVE",
		[20] = "ADA",
		[21] = "LISP",
		[22] = "RUBY",
		[23] = "EIFFEL",
		[24] = "EIFFELKW",
		[25] = "TCL",
		[26] = "NNCRONTAB",
		[27] = "BULLANT",
		[28] = "VBSCRIPT",
		[29] = "ASP",
		[30] = "PHP",
		[31] = "BAAN",
		[32] = "MATLAB",
		[33] = "SCRIPTOL",
		[34] = "ASM",
		[35] = "CPPNOCASE",
		[36] = "FORTRAN",
		[37] = "F77",
		[38] = "CSS",
		[39] = "POV",
		[40] = "LOUT",
		[41] = "ESCRIPT",
		[42] = "PS",
		[43] = "NSIS",
		[44] = "MMIXAL",
		[45] = "CLW",
		[46] = "CLWNOCASE",
		[47] = "LOT",
		[48] = "YAML",
		[49] = "TEX",
		[50] = "METAPOST",
		[51] = "POWERBASIC",
		[52] = "FORTH",
		[53] = "ERLANG",
		[54] = "OCTAVE",
		[55] = "MSSQL",
		[56] = "VERILOG",
		[57] = "KIX",
		[58] = "GUI4CLI",
		[59] = "SPECMAN",
		[60] = "AU3",
		[61] = "APDL",
		[62] = "BASH",
	}
	print(lexers[editor.Lexer])
--~ 	RepeatEditor("LineUp", 4)
--~ 	editor:Home()
--~ 	RepeatEditor("LineDownExtend", 2)

--~ 	local e = editor
--~ 	local lu = e.LineUp
--~ 	local lde = e.LineDownExtend
--~ 	lu(e); lu(e); lu(e); lu(e)
--~ 	editor:Home()
--~ 	lde(e); lde(e)

--~ 	editor:Clear()
--~ 	editor:LineDown()
--~ 	editor:LineDownExtend(); editor:LineDownExtend(); editor:LineDownExtend(); editor:LineDownExtend(); editor:LineDownExtend()
--~ 	editor:Clear()
end

--------------------------------------------------------------------------------
--- Event Handlers ---

--================ Super Abbreviations ================
local abbrList =
{
	["Infinite Loop"] =
	{
		template = [[
for (;;)
{
	|;
}
]],
		posChar = '|'
	},
	["While Loop"] =
	{
		template = [[
while (|)
{
	;
}
]],
		posChar = '|'
	},
	["gsub"] =
	{
		template = [[
string.gsub(s, "^",
	function (s)
		return s
	end)
]],
		posChar = '^'
	},
}
local titles = ''
for t, v in pairs(abbrList) do
	titles = titles .. t .. '~'
end
titles = string.sub(titles, 1, -2)

function SuperAbbr()
	editor.AutoCSeparator = 126
	editor:UserListShow(10, titles)
end

function OnUserListSelection(lt, txt)
	if lt == 10 then
		editor:InsertText(-1, abbrList[txt]["template"])
		-- Should locate posChar, remove it and put the caret there
		return true
	else
		return false
	end
end

--=====================================================

--~ function OnSwitchFile(filename)
--~ 	print("Lua sample: switched to " .. filename)
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnOpen(filename)
--~     if string.sub(filename, -1) ~= '\\' then
--~         print("Lua sample: opened " .. filename)
--~     end
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnSave(filename)
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnBeforeSave(filename)
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnChar(charAdded)
--~ 	return false	-- Let next handler to process event
--~ end

--## Real time conversion of (some) high Ascii characters to their entities.
-- I don't use it because I use the right encoding (ISO-8859-1) so 8bit is fine.

local entities =
{
--	['&'] = "&amp;",
--	['<'] = "&lt;",
--	['>'] = "&gt;",
	-- French entities (the most common ones)
	[''] = "&agrave;",
	[''] = "&acirc;",
	[''] = "&eacute;",
	[''] = "&egrave;",
	[''] = "&ecirc;",
	[''] = "&euml;",
	[''] = "&icirc;",
	[''] = "&iuml;",
	[''] = "&ocirc;",
	[''] = "&ouml;",
	[''] = "&ugrave;",
	[''] = "&ucirc;",
	[''] = "&yuml;",
	[''] = "&Agrave;",
	[''] = "&Acirc;",
	[''] = "&Eacute;",
	[''] = "&Egrave;",
	[''] = "&Ecirc;",
	[''] = "&Euml;",
	[''] = "&Icirc;",
	[''] = "&Iuml;",
	[''] = "&Ocirc;",
	[''] = "&Ouml;",
	[''] = "&Ugrave;",
	[''] = "&Ucirc;",
	[''] = "&ccedil;",
	[''] = "&Ccedil;",
	[''] = "&Yuml;",
	[''] = "&laquo;",
	[''] = "&raquo;",
	[''] = "&copy;",
	[''] = "&reg;",
	[''] = "&aelig;",
	[''] = "&AElig;",
	[''] = "&OElig;", -- Not understood by all browsers - Alternative: &#338;
	[''] = "&oelig;", -- Not understood by all browsers - Alternative: &#339;
}

local htmlExt = { htm = 1, html = 1, php = 1, asp = 1 }

function On_Char(charAdded)
	if htmlExt[string.lower(props["FileExt"])] == 1 then	-- Only when editing an HTML file...
		if entities[charAdded] ~= nil then
--			trace(charAdded)
			local pos = editor.CurrentPos
			editor:SetSel(pos, pos - 1)
			editor:ReplaceSel(entities[charAdded])
		end
	end
	return false	-- Let next handler to process event
--~ 	return true	-- Don't let next handler to process event
end

--## Autoclose braces and quotes.
-- This is mostly annoying, I should try the Eclipse way:
-- add the closing match, but if it is typed anyway, eat it.
-- Still annoying in VB where a single quote is used for line comments...

local toClose = { ['('] = ')', ['{'] = '}', ['['] = ']', ['"'] = '"', ["'"] = "'" }

function On__Char(charAdded)
--	trace(charAdded)

	if toClose[charAdded] ~= nil then
		local pos = editor.CurrentPos
		editor:ReplaceSel(toClose[charAdded])
		editor:SetSel(pos, pos)
	end
	return false	-- Let next handler to process event
--	return true	-- Don't let next handler to process event
end

--## Put SQL keywords on uppercase while typing them.

local sqlExt = { sql = 1, plsql = 1, mssql = 1, foosql = 1 }
local sqlKW =
{
absolute=1, action=1, add=1, admin=1, after=1, aggregate=1,
alias=1, all=1, allocate=1, alter=1, ['and']=1, -- Use [''] because it is a Lua keyword...
any=1, are=1, array=1, as=1, asc=1,
assertion=1, at=1, authorization=1,
}

function GetPreviousWord()
	local pos = editor.CurrentPos
	editor:WordLeftExtend()
	local word = editor:GetSelText()
	editor:CharRight()
	return word
end

function On__Char(charAdded)
	if sqlExt[string.lower(props["FileExt"])] == 1 then	-- Only when editing an SQL file...
		local word = GetPreviousWord()
		if sqlKW[word] == 1 then
--			trace(charAdded)
			local pos = editor.CurrentPos
			editor:SetSel(pos, pos - string.len(word))
			editor:ReplaceSel(string.upper(word))
		end
	end
	return false	-- Let next handler to process event
--~ 	return true	-- Don't let next handler to process event
end

--~ function OnSavePointReached()
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnSavePointLeft()
--~ 	return false	-- Let next handler to process event
--~ end

--~ function OnDoubleClick()
--~ 	return 0	-- Let next handler to process event
--~ end

--~ function OnMarginClick()
--~ 	return 0	-- Let next handler to process event
--~ end

-- Example styler looks for "lua" and highlights to end of word.
function OnStyle(start, len, state, styler)
	--trace("> [" .. start .."," .. len .. "," .. startstyle .. "]\n")
	local i = start
    while i <= start + len do
		if string.lower(editor:GetRange(i, i+3)) == "lua" then
			styler:ColourTo(i - 1, state)
			state = 1
		elseif state == 1 then
			if not string.find(editor:GetRange(i, i+1), "[%a_]") then
				styler:ColourTo(i - 1, state)
				state = 0
			end
		end
		i = i + 1
	end
	styler:ColourTo(start + len - 1, state)
	return 1	-- Event have been handled, don't give it to default processing
end
